from itertools import groupby, product

from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.shape import get_shape, get_box

Line = tuple[Pixel, Pixel]


def get_line(pixel: Pixel, image: SimpleImage) -> tuple[Line, list[Pixel], int] | None:
    pixels = get_shape(pixel, image)
    if len(set(pixels.values())) != 1:
        return None
    (x_min, y_min), (right, bottom), _ = get_box(pixels)
    x_max = right - 1
    y_max = bottom - 1
    # \ or /
    if (x_min, y_min) in pixels and (x_max, y_max) in pixels:
        end1 = Pixel(x_min, y_min)
        end2 = Pixel(x_max, y_max)
        forward = True
    elif (x_min, y_max) in pixels and (x_max, y_min) in pixels:
        end1 = Pixel(x_min, y_max)
        end2 = Pixel(x_max, y_min)
        forward = False
    else:
        return None
    if end1 == end2:
        return None
    if image.is_on_edge(end1) and image.is_on_edge(end2):
        return None
    on_edge = image.is_on_edge(end1) or image.is_on_edge(end2)
    if not on_edge and (end1.x - end2.x)**2 + (end1.y - end2.y)**2 < 8:
        return None
    sorted_pixels = sorted(pixels)
    slices = [list(y for x, y in group) for key, group in groupby(sorted_pixels, key=lambda px: px.x)]
    # all slices are single intervals
    for s in slices:
        if not all(y1 + 1 == y2 for y1, y2 in zip(s, s[1:])):
            return None
    # slices are monotonic
    for slice1, slice2 in zip(slices, slices[1:]):
        if forward and (slice1[0] > slice2[0] or slice1[-1] > slice2[-1]):
            return None
        if not forward and (slice1[0] < slice2[0] or slice1[-1] < slice2[-1]):
            return None
    neighbor_counts = set()
    for pixel in pixels:
        x, y = pixel
        neighbor_counts.add(
            sum(1 for i, j in product(range(x - 1, x + 2), range(y - 1, y + 2)) if (i, j) in pixels)
        )
    if on_edge or max(neighbor_counts) == 3 or (min(neighbor_counts) >= 2 and 6 <= max(neighbor_counts) <= 8):
        width = 1 if max(neighbor_counts) == 3 else 2
        return (end1, end2), sorted_pixels, width



from functools import cache
from logging import getLogger
from typing import Iterable, NamedTuple

from PIL import Image

from parse_qwantz.colors import Color, square_distance, get_color_with_threshold, WHITE

logger = getLogger()


class Pixel(NamedTuple):
    x: int
    y: int


class UnknownColor(ValueError):
    pass


def get_pixels(image: Image) -> Iterable[tuple[Pixel, Color]]:
    palette = image.getpalette()
    palette = tuple(palette) if palette else None
    values = image.getdata()
    width = image.width
    unknown_colors = False
    for i, value in enumerate(values):
        x = i % width
        y = i // width
        value = normalize_color(value, palette)
        if value != WHITE:
            color = get_color_with_threshold(value)
            if not color:
                if not unknown_colors:
                    unknown_colors = True
                    logger.warning(f"Unknown color at {(x, y)}: {value}.")
                color = Color(*value)
            if color != WHITE:
                yield Pixel(x, y), color


@cache
def normalize_color(
    color: int | tuple[int, int, int] | tuple[int, int, int, int], palette: tuple[int, ...] | None
) -> tuple[int, int, int]:
    if palette is None and isinstance(color, int):
        color = (color, color, color)
    elif palette is None and isinstance(color, tuple) and len(color) == 4:
        color = color[:3]
    elif palette is not None:
        color = palette[color * 3: color * 3 + 3]
    return color


def remove_subsequence(sorted_pixels: list[Pixel], subsequence: list[Pixel]) -> list[Pixel]:
    sub_iter = iter(subsequence)
    next_pixel = next(sub_iter)
    pixels = []
    for pixel in sorted_pixels:
        while next_pixel is not None and pixel > next_pixel:
            try:
                next_pixel = next(sub_iter)
            except StopIteration:
                next_pixel = None
        if pixel != next_pixel:
            pixels.append(pixel)
    return pixels


def is_ask_professor_science(image: Image) -> bool:
    palette = image.getpalette()
    palette = tuple(palette) if palette else None
    for x in range(109, 113):
        for y in range(1, 6):
            pixel_color = normalize_color(image.getpixel((x, y)), palette)
            if square_distance(pixel_color, (224, 231, 248)) <= 3:
                return True
            if square_distance(pixel_color, (209, 220, 244)) <= 3:
                return True
    return False



import logging
import sys


class ColorFormatter(logging.Formatter):
    GREY = "\x1b[38;20m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m"
    BOLD_RED = "\x1b[31;1m"
    RESET = "\x1b[0m"
    FORMAT = "%(levelname)s:%(panel)s %(message)s"

    FORMATS = {
        logging.DEBUG: GREY + FORMAT + RESET,
        logging.INFO: GREY + FORMAT + RESET,
        logging.WARNING: YELLOW + FORMAT + RESET,
        logging.ERROR: RED + FORMAT + RESET,
        logging.CRITICAL: BOLD_RED + FORMAT + RESET
    }

    def __init__(self, *args, colors: bool = True, defaults=None, **kwargs):
        self._defaults = defaults or {"panel": ""}
        self._colors = colors
        super().__init__(*args, defaults=defaults, **kwargs)

    def format(self, record):
        if sys.stderr.isatty() and self._colors:
            log_fmt = self.FORMATS.get(record.levelno)
        else:
            log_fmt = self.FORMAT
        formatter = logging.Formatter(log_fmt, defaults=self._defaults)
        return formatter.format(record)


def set_logging_formatter():
    logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setFormatter(ColorFormatter())
    logger.addHandler(handler)
